#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__title__ = 'readmet'
__description__ = ('Python library for reading and writing '
                   'less popular formats used in meteorology.')
__url__ = ''
__version__ = '0.8.26'
__author__ = u'Clemens Drüe'
__author_email__ = 'druee@uni-trier.de'
__license__ = 'EUPL-1.2'
__copyright__ = '(C) 2019-2025 Clemens Drüe'
